/* This is a driver for a wacom graphics tablet.
 * It was based on my PCIbase driver, from
 *  http://www.bolthole.com/solaris/drivers/
 *
 * It is being converted to a STREAMS type driver now.
 *    Philip Brown
 *
 *  @(#) wacom.c 1.22@(#)
 */

#include <varargs.h>
#define _SYS_VARARGS_H		/* sys/varargs.h hack for gcc */

#include <sys/usb/usba.h>
  /* we dont actually use the hid routines.. we use the defines */
#include <sys/usb/clients/hid/hid.h>

#include <sys/errno.h>
#include <sys/conf.h>
#include <sys/modctl.h>
#include <sys/open.h>
#include <sys/stat.h>
#include <sys/stream.h>
#include <stropts.h>
#include <sys/signal.h>

/* THESE MUST BE LAST */
#include <sys/ddi.h>
#include <sys/sunddi.h>


/* This value goes in the USB "policy" info struct */
#define WACOM_MAX_TRANSFER 10
#define WACOM_BUFFER_SIZE 512


#undef DO_JITTER

/* As well as changing the NAME of this, you will also want to change
 * the TYPE of the minor node created.
 */
#define DEVNODE_NAME	"wacom"	
#define DEVNODE_TYPE	DDI_NT_SERIAL /* see sunddi.h */


/**********************************************************************/

/* This ASSUMES you create minor nodes where the higher bits denote
 * the instance of the driver. (aka "which card is this?")
 * The lower bits denote clone number, or anything else you want.
 */

#define DEVT_TO_INST(devt) (getminor(devt)>>4)
#define DEVT_TO_CLONENUM(devt) (getminor(devt) & 0xf)

/* Used to choose a node  num for a device node compatible with DEVT_TO_INST*/
#define INST_TO_NODENUM(instance) (instance<<4)


/*Our globally visible stuffs, that we populate the global
 * driver entry structs with  [base_cb_ops, base_ops]
 * Yeah, its kind of wierd to have a "static" be also externally visible,
 * but that's traditionally how these are declared.
 */
static	int	wacom_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg,
		    void **result);

static	int	wacom_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);

static	int	wacom_open(queue_t *q,dev_t *dev, int ofl, int otyp, cred_t *c);
static	int	wacom_wput(queue_t *q,mblk_t *mblk);

static	int	wacom_rput(queue_t *q,mblk_t *mp);
static	int	wacom_rsrv(queue_t *q);

static	int	wacom_close(queue_t *q, int flag, cred_t *cred_p);

static	int	wacom_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);


static int debug_level;


/**********************************************************************/
/* STREAMS driver info stuff  */

static struct module_info stream_minfo = {
	0x0001,		/* module id number (??) */
	"wacom",	/* module name */
	0,		/* min packet size */
	INFPSZ,		/* max packet size */
	512,		/* "high water mark" (??)*/
	128,		/* "low water mark" (??)*/
};

/* Queue for reading data */
static struct qinit stream_rinit = {
	wacom_rput, /* qi_putp */
	wacom_rsrv, /* qi_srvp */
	wacom_open,
	wacom_close,
	NULL, /* qi_qadmin */
	&stream_minfo,
	NULL /* qi_mstat */
};

/* Queue for writing data (to the dev) */
static struct qinit stream_winit = {
	/* Im not sure we should really have ANYTHING here.. */
	wacom_wput, /* qi_put */ 
	NULL, /* qr_srv not needed */
	NULL, /* qi_qopen */
	NULL, /* qi_qclose */
	NULL, /* qi_qadmin */
	&stream_minfo,
	NULL /* qi_mstat */
};

static struct streamtab stream_rwinfo = {
	&stream_rinit,
	&stream_winit,
	NULL,		/* st_muxrinit */
	NULL		/* st_muxwinit */
};

/* end of STREAMS specific struct stuff  */
/**********************************************************************/



/* "char/block operations" */
/* Since we are a STREAMS dev, all the stuff that is NORMALLY referenced in 
 * here, is now referenced from the wierd STREAMs specific subtree.
 */
static struct cb_ops	wacom_cb_ops = {
	nulldev,		/* open() */
	nulldev,		/* close() */
	nodev,			/* strategy()		*/
	nodev,			/* print routine	*/
	nodev,			/* no dump routine	*/
	nodev,			/* read() */
	nodev,			/* write() */
	nodev,			/* ioctl */
	nodev,			/* no devmap routine	*/
	nodev,			/* no mmap routine	*/
	nodev,			/* no segmap routine	*/
	nochpoll,		/* no chpoll routine	*/
	ddi_prop_op,
	&stream_rwinfo,		/* cb_stream struct */
	D_NEW | D_MP,		/* safe for multi-thread/multi-processor */
	CB_REV,			/* cb_ops version? */
	nodev,			/* cb_aread() */
	nodev,			/* cb_awrite() */
};

/* "device operations". */
static struct dev_ops wacom_ops = {
	DEVO_REV,		/* devo_rev */
	0,			/* devo_refcnt */
	wacom_getinfo,		/* devo_getinfo */
	nulldev,		/* devo_identify */
	nulldev,		/* devo_probe */
	wacom_attach,		/* devo_attach */
	wacom_detach,		/* devo_detach */
	nodev,			/* devo_reset */
	&wacom_cb_ops,		/* devo_cb_ops */
	(struct bus_ops *)0,	/* devo_bus_ops */
	NULL,			/* devo_power */
};

/* Ignore this.
 * It is already declared in <sys/modctl.h>
 * It is only a filler for the next struct.
 */
extern	struct	mod_ops mod_driverops;

static	struct modldrv modldrv = {
	&mod_driverops,
	"BOLTwacom STREAM driver",
	&wacom_ops,
};

static	struct modlinkage modlinkage = {
	MODREV_1,		/* MODREV_1 is indicated by manual */
	{&modldrv,NULL,NULL,NULL}

};


/************************************************************
 * And here's finally the vars specific to THIS driver,
 * that get allocated on a per-instance basis.
 * Here's where you should keep instance-specific mutexes,
 * DMA handles, register mapping handles, and stuff like that
 *
 ************************************************************/


#define MAX_INSTANCES 4 /* max of 4 hardware devices loaded */

/* 
 * Note that this mutex gets used in _init. Which means this mutex
 * isn't just for this "instance" of the driver. It is for ALL INSTANCES.
 * If you have multiple cards, you have multiple instances of the driver,
 * with this single shared mutex across them all
 */
static kmutex_t   wacom_global_mutex;

/* This is SORT OF like	int (*array)[] , except you NEVER USE THIS DIRECTLY.
 * You just use it with ddi_get_soft_state().
 * 
 *    "Global Softstate Handle" 
 */
void * wacom_glob_soft_handle;

/* This gets filled out on a per-instance basis by wacom_attach */
struct wacom_softstate {
	dev_info_t *dip; /* you always need this cached */
	kmutex_t instance_mutex;
	kcondvar_t read_cv;

	/* ways to squish "jitter". eg: mouse hasnt moved, but
	 * we still get phantom damn events from the hardware
	 */
	int jitter_bits;
	uint8_t lastdata[8];

	usb_config_descr_t usb_config;
	usb_endpoint_descr_t endpoint;
	int usb_interfacenum;
	usb_interface_descr_t usb_interface_d;
	int openflags; /* tell if we have been open()ed or not */
	usb_pipe_handle_t usb_pipe_handle;

	size_t max_transfer_size; /* technically, BULK max transfer... */

	queue_t *rq, *wq;

	/* since we are not a streams driver we have to make our
	 * own buffer/queue
	 * "end" is where to add new data. "start" is where read()
	 * should take data from
	 */
	unsigned char *readbuffer;
	unsigned char *readbuf_start,*readbuf_end;

};

/* defines used on the "openflags" var in softstate*/
#define WACOM_OPEN_HELD		0x01
#define WACOM_OPEN_ENDING	0x02
#define WACOM_OPEN_SHUTDOWN	0x03 /* flag "shutting down" case */


/*****************************************************************/
/*****************************************************************/
/*              Start of actual kernel routines here             */
/*****************************************************************/
/*****************************************************************/



void
wacom_debugprint(va_alist)
va_dcl
{
	va_list args;
	char *fmt;
	if(debug_level==0){
		return;
	}

	va_start(args);
	fmt = va_arg(args, char *);
	/* DO NOT USE cmn_err here!!! it will not be interrupt-safe */
	vprintf(fmt, args);
	va_end(args);
}



/*Like the wacom_global_mutex above, this routine is GLOBAL across all
 * instances.
 * It actually gets called when you add_drv your driver from the command
 * line, before  * it is hooked up to any specific hardware.
 * See wacom_attach() below for instance-specific stuff.
 *
 * HOWEVER... Its often nice to pre-allocate "softstate" space
 * here, if you dont mind hardcoding the total number of possible
 * instances you can have. So we do that.
 * 
 */
int
_init(void)
{
       int     ret;

	printf("wacom: _init called\n");
	/*
               * Initialize global mutex before mod_install'ing driver.
               * If mod_install() fails, must clean up mutex initialization
               */
	mutex_init(&wacom_global_mutex, "wacom: Global Mutex",
                     MUTEX_DRIVER, (void *)NULL);

	/* This is kind of like
	 * malloc(sizeof(struct wacom_softstate *MAX_INSTANCES))
	 */
	ret=ddi_soft_state_init(&wacom_glob_soft_handle,
				sizeof(struct wacom_softstate),
			        MAX_INSTANCES);
	if(ret!=0){
		return (ret);
	}
	
	if ((ret = mod_install(&modlinkage)) != 0) {
		mutex_destroy(&wacom_global_mutex);
		ddi_soft_state_fini(&wacom_glob_soft_handle);
		cmn_err(CE_WARN, "wacom: _init: mod_install FAILED:\n");
		return DDI_FAILURE;
	}

	wacom_debugprint("wacom: _init returning SUCCESS\n");
	return (DDI_SUCCESS);
}

int
_info(struct modinfo *modinfop)
{
	
	wacom_debugprint("wacom: _info called\n");
	return (mod_info(&modlinkage, modinfop));
}

/* This is the match to the _init() routine. Clean up stuff allocated
 * specifically in _init() here.
 * Stuff in attach() should be cleaned up in detach()
 */
int
_fini(void)
{
	int ret;

	wacom_debugprint("wacom: _fini called\n");
	
        /*
         * If mod_remove() is successful, we destroy our global mutex
         */
	if ((ret = mod_remove(&modlinkage)) == 0) {
		ddi_soft_state_fini(&wacom_glob_soft_handle);
		mutex_destroy(&wacom_global_mutex);
	}
	return (ret);
}

/************************************************************/

#ifdef DO_JITTER
/* returns true if data is just 'jitter', or false (0) if 'real' data.
 * we do not modify the data
 *
 * This routine assumes wacom graphire2 data format,
 *  data[]=={??, ??, xLSB,xMSB,yLSB,yMSB,pLSB,pMSB}
 *
 * Trouble is... this routine doesnt really work very well,
 * algorythmically speaking. so I dont use it.
 */

static int isjitterdata(struct wacom_softstate *softstate,uint8_t *data)
{
	uint8_t jitbits;

	if(data[3] !=softstate->lastdata[3]) return 0;
	if(data[5] !=softstate->lastdata[5]) return 0;
	if(data[7] !=softstate->lastdata[7]) return 0;

	jitbits=softstate->jitter_bits & 0xff;
	printf("Doing compare: %x-%x %x-%x %x-%x: jit=%x (%x)\n",
	       (data[2]|jitbits), (softstate->lastdata[2]|jitbits),
	       (data[4]|jitbits), (softstate->lastdata[4]|jitbits),
	       (data[6]|jitbits), (softstate->lastdata[6]|jitbits),
	       jitbits, softstate->jitter_bits
	       );

	if((data[2]|jitbits) != (softstate->lastdata[2]|jitbits)) return 0;
	if((data[4]|jitbits) != (softstate->lastdata[4]|jitbits)) return 0;
	if((data[6]|jitbits) != (softstate->lastdata[6]|jitbits)) return 0;

	/* no significant difference from last data received.
	 * consider it "jitter". return true.
	 */
	return 1;
}
#endif

/**********************************************************************/
/* This is probably interrupt context. 
 * mblk_t is a STREAMS style Message Block. Arrg.
 * Right now, I can only test Graphire tablets, since that is what
 * I HAVE. The Intuos has a different data message sent.
 */
static int
wacom_pipe_callback(usb_pipe_handle_t pipe, usb_opaque_t arg,mblk_t*mblkptr)
{
	struct wacom_softstate *softstate=(struct wacom_softstate*)arg;

	if(debug_level>1){
		wacom_debugprint("wacom_pipe_callback called\n");
	}

	if(mblkptr==NULL){
		printf("mblk==NULL\n");
		return (USB_SUCCESS);
	}

	if(softstate->rq==NULL){
		printf("ERROR in callback: softstate->rq==NULL?\n");
		return USB_FAILURE;
	}

	wacom_debugprint("^wacom_pipe_callback - calling wacom_rput\n");
	wacom_rput(softstate->rq, mblkptr);

	return (USB_SUCCESS);
}

/* temp test routine, that is only used for test purposes, essentially.
 */
static void
wacom_test_callback(usb_opaque_t arg,int arg1, uint_t arg2)
{
	wacom_debugprint("wacom_TEST_callback called\n");
	return;
}

/* This gets called for partial USB transfers.
 * EG: some error occurred.
 */
static int
wacom_exception_callback(usb_pipe_handle_t pipe, usb_opaque_t arg,
			 uint_t complete_reason,mblk_t*mblkptr,uint_t flag)
{
	/*struct wacom_softstate *softstate=(struct wacom_softstate*)arg;*/

	wacom_debugprint("wacom_EXCEPTION_callback called: reason %d\n",
			 complete_reason);
	return (USB_SUCCESS);
}

static int wacom_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd,
			     void *arg, void **result){
	struct wacom_softstate *softstate;
	int returnval=DDI_FAILURE;
	int instance=getminor((dev_t)arg);
	
	wacom_debugprint("wacom: getinfo called\n");
	
	switch(infocmd){
	    case DDI_INFO_DEVT2DEVINFO:
		softstate=ddi_get_soft_state(wacom_glob_soft_handle,instance);
		if(softstate==NULL){
			*result=NULL;
			break;
		}
		*result=softstate->dip;
		returnval=DDI_SUCCESS;
		break;
	    case DDI_INFO_DEVT2INSTANCE:
		*result=(void*)instance;
		returnval=DDI_SUCCESS;
		break;
	}

	return (returnval);
}




/* This routine can be called by wacom_attach (if errors occur)
 * and wacom_detach (on normal detach)
 * Idea is to clean up any mutexes, register mappings, etc. nicely.
 * We also free the instance-softstate.
 * We do NOT remove the minor node, however. Only detach does
 * that currently.
 */
static void wacom_cleanup(struct wacom_softstate *softstate){
	int instance;

	if(softstate==NULL){
		/* yes, paranoia is good*/
		cmn_err(CE_WARN,"ERROR: wacom_cleanup called with NULL");
		return;
	}
   	instance=ddi_get_instance(softstate->dip);

	if(softstate->readbuffer!=NULL){
		kmem_free(softstate->readbuffer,WACOM_BUFFER_SIZE);
		mutex_destroy(&softstate->instance_mutex);
		cv_destroy(&softstate->read_cv);
	}

	ddi_soft_state_free(wacom_glob_soft_handle,instance);
}


#define REQUEST_TYPE USB_DEV_REQ_RECIPIENT_INTERFACE|USB_DEV_REQ_TYPE_CLASS
#define REQUEST_INDEX softstate->usb_interface_d.bInterfaceNumber
#define REQUEST_CMD USB_REQ_SET_CONFIGURATION

/*
 * set_dev changes the "report type" of the tablet, from the default "type 1",
 * to "type 2". This gives us full info on pressure, and absolute positioning.
 *
 *In an ideal world, I would use the "hid" interfaces to do this
 * sort of thing. But since that isnt documented, and since "SET_REPORT"
 * is really just a hid layer on top of a "regular" USB call...
 * Do the "regular call" directly. It should get through whether we
 * are on top of hid, or not.
 */
void set_devstate(struct wacom_softstate *softstate){
	usb_pipe_handle_t pipe_handle;
	uint_t failreason;
	mblk_t *mblkptr;

	if(usb_pipe_open(softstate->dip,
			 NULL,NULL,USB_FLAGS_OPEN_EXCL,
			 &pipe_handle) != USB_SUCCESS){
		cmn_err(CE_WARN,"set_dev: usb_pipe_open failed");

	}
	wacom_debugprint("set_dev: secondary pipe_open suceeded! Closing now.\n");
	wacom_debugprint(" Note: interfacenum = %d,endpoint=%d\n",
			 REQUEST_INDEX,usb_endpoint_num(pipe_handle));

	mblkptr=allocb(2,0);
	if(mblkptr==NULL){
		cmn_err(CE_WARN,"set_dev: allocb failed!");
		usb_pipe_close(&pipe_handle,USB_FLAGS_SLEEP,NULL,NULL);
		return;
	}
	
	mblkptr->b_rptr[0]=2;
	mblkptr->b_rptr[1]=2;
	mblkptr->b_wptr+=2;
	wacom_debugprint(" set_dev: MBLK data ptr==%p, db_lib=%p\n",
			mblkptr->b_datap->db_base,
			mblkptr->b_datap->db_lim);


#ifdef RESETIT
	if(usb_pipe_sync_device_ctrl_send(pipe_handle,
			USB_DEV_REQ_RECIPIENT_ENDPOINT,
			USB_REQ_CLEAR_FEATURE,
			USB_ENDPOINT_HALT,
			usb_endpoint_num(pipe_handle),
			0, /*wLength*/
			NULL,
			&failreason,
			USB_FLAGS_ENQUEUE) != USB_SUCCESS){
		wacom_debugprint("set_dev: CLEAR_FEATURE failed\n");
		wacom_debugprint("  failreason=%d\n",failreason);

	} else {
		wacom_debugprint("set_dev: clear_feature SUCCEEDED!\n");
	}
#endif

	if(usb_endpoint_num(pipe_handle) !=0){
		cmn_err(CE_WARN,"ERROR: set_dev does not have endpoint 0");
		usb_pipe_close(&pipe_handle,USB_FLAGS_SLEEP,NULL,NULL);
		return;
	}

	/* hardcoding SET_CONFIGURATION from values from Dave F.
	 */

	if(usb_pipe_sync_device_ctrl_send(pipe_handle,
			REQUEST_TYPE,
			HID_SET_REPORT,
			REPORT_TYPE_FEATURE |2, /* wValue */
			softstate->usb_interfacenum, /* wIndex, probably 0*/
			2, /*wLength*/
			mblkptr, /* mblk data */
			&failreason,
			USB_FLAGS_ENQUEUE) != USB_SUCCESS){
		wacom_debugprint("set_dev: SET_CONFIGURATION FAILED! \n");
		wacom_debugprint("  failreason=%d\n",failreason);

		/* Apparently, if I try to free this, the system blows up.
		 * So I wont try!
		 */
		/*freemsg(mblkptr);*/
	} else {
		wacom_debugprint("set_dev: SET_CONFIGURATION=2 succeeded! \n");
	}

	usb_pipe_close(&pipe_handle,USB_FLAGS_SLEEP,NULL,NULL);

}

/*
 * _init() has already been called. This routine gets called
 * once for EVERY MATCHING PIECE OF HARDWARE.
 * It gets called for initial "attach", and also for resume(/suspend)
 * You would usually allocate space for state variables, and
 * map device registers in here.
 *
 * To avoid reboot loops, be very very careful putting things in here.
 * If you want to put extra debug type stuff in, it's safer putting
 * it in wacom_open() or someplace like that, that gets called after
 * the driver is safely loaded
 *
 * Be sure to free/deallocate anything you create here, in
 *  wacom_detach()
 */
static int wacom_attach(dev_info_t *dip, ddi_attach_cmd_t cmd){
   	int instance;
	struct wacom_softstate *softstate;
	size_t usb_raw_len;
	uchar_t *usb_rawconf;

	/* default debug level to **0**
	 * You can override in the .conf file.
	 */
	debug_level=ddi_prop_get_int(DDI_DEV_T_ANY,dip, DDI_PROP_DONTPASS,
                                    "debug-level", 0);

	wacom_debugprint(" wacom_attach called\n");

   	switch(cmd){
	 case DDI_RESUME:
	   	cmn_err(CE_NOTE, "wacom: Ignoring attach_RESUME\n");
		return DDI_FAILURE;
	 case DDI_PM_RESUME:
		cmn_err(CE_NOTE, "wacom: Ignoring attach_PM_RESUME\n");
		return DDI_FAILURE;
	 case DDI_ATTACH:
	   	break;
	 default:
	   	cmn_err(CE_NOTE, "wacom: unknown attach command\n");
	   	return DDI_FAILURE;
	}
   	instance=ddi_get_instance(dip);
   	wacom_debugprint("wacom: attach for instance %d\n", instance);


	/* We dont have to check for being over MAX_INSTANCES, because
	 * these two calls do it implicitly
	 */
	if (ddi_soft_state_zalloc(wacom_glob_soft_handle, instance) != DDI_SUCCESS) {
		cmn_err(CE_WARN, "wacom: %d: Failed to alloc soft state",
		    instance);
		return (DDI_FAILURE);
	}
	softstate=ddi_get_soft_state(wacom_glob_soft_handle,instance);
	if(softstate==NULL){
		cmn_err(CE_WARN, "wacom: attach for instance %d: Failed to get soft state",
			instance);
		ddi_soft_state_free(wacom_glob_soft_handle,instance);
		return (DDI_FAILURE);
	}

	/* Normal resolution of coordiates is 10000x10000.
	 * The device doesnt even support that much res!!
	 * So throw away low-end bits as "jitter"
	 */
	softstate->jitter_bits=ddi_prop_get_int(DDI_DEV_T_ANY,dip,
						DDI_PROP_DONTPASS,
						"jitter-bits", 0xf);
	softstate->jitter_bits &= 0xff;

	wacom_debugprint("WACOM: jitter_bits=%x\n",softstate->jitter_bits);
	softstate->dip = dip;
	softstate->openflags = 0; /* redundant, but what the heck */
	


	mutex_init(&softstate->instance_mutex, NULL, MUTEX_DRIVER, NULL);
	cv_init(&softstate->read_cv, NULL, CV_DRIVER,NULL);

	softstate->readbuffer=kmem_zalloc(WACOM_BUFFER_SIZE,KM_SLEEP);
	/* the buffer pointers get cleared in wacom_open() */


	usb_rawconf = usb_get_raw_config_data(dip,&usb_raw_len);

	/* Fill out "configuration" descriptor */
	if(usb_parse_configuration_descr(usb_rawconf, usb_raw_len,
					&softstate->usb_config,
					USB_CONF_DESCR_SIZE) !=
						USB_CONF_DESCR_SIZE)
	{
		cmn_err(CE_WARN,"usb_parse_configuration_desc failed\n");
		wacom_cleanup(softstate);
		return DDI_FAILURE;
	}
	wacom_debugprint("USB config num is %d\n",
			 softstate->usb_config.bConfigurationValue);
	softstate->usb_interfacenum=usb_get_interface_number(dip);
	wacom_debugprint("USB interface num is %d\n",
				softstate->usb_interfacenum);

	/* If interfacenum==-1, set it to 0: we are responsible for full dev*/
	if(softstate->usb_interfacenum==-1){
		wacom_debugprint("  (resetting interface num to 0)\n");
		softstate->usb_interfacenum=0;
	}

	/* Fill out "interface" descriptor */
	if(usb_parse_interface_descr(usb_rawconf,usb_raw_len,
				    softstate->usb_interfacenum,
				    0,
				    &softstate->usb_interface_d,
				    USB_IF_DESCR_SIZE) != USB_IF_DESCR_SIZE)
	{
		cmn_err(CE_WARN,"usb_parse_interface_desc failed");
		wacom_cleanup(softstate);
		return DDI_FAILURE;
	}

	wacom_debugprint("number of endpoints = %d\n",
			 softstate->usb_interface_d.bNumEndpoints);

	/* Fill out "endpoint" descriptor */
	if(usb_parse_endpoint_descr(usb_rawconf, usb_raw_len,
					softstate->usb_interfacenum,
					0,/* alt interfacenum*/
					0,/*endpoint index num*/
					&softstate->endpoint,
					USB_EPT_DESCR_SIZE) !=
						USB_EPT_DESCR_SIZE)
	{
		cmn_err(CE_WARN,"usb_parse_endpoint_descr failed\n");
		wacom_cleanup(softstate);
		return DDI_FAILURE;
	}

	/* FYI: type3==USB_EPT_ATTR_INTR */
	wacom_debugprint("endpoint %d: type %d",
			 0,
			 softstate->endpoint.bmAttributes);

	wacom_debugprint("INTERFACE Class=%x, subclass=%x\n",
			 softstate->usb_interface_d.bInterfaceClass,
 			 softstate->usb_interface_d.bInterfaceSubClass);

	if(usb_pipe_bulk_transfer_size(dip,&softstate->max_transfer_size)!=USB_SUCCESS)
	{
		cmn_err(CE_NOTE,"wacom: bulk_transfer_size detect failed");
	}
	wacom_debugprint(" bulk_maxtransfer sizeval=%u",softstate->max_transfer_size);
	wacom_debugprint(" maxtransfer sizeval=%u",
	                  softstate->endpoint.wMaxPacketSize);


#ifdef STRINGSAMPLE
	/* We dont need to do this, because ddi_dev_report() does
	 * the same thing. See /var/adm/messages, Not console.
	 * But this is here in case you want an
	 * example of pulling string info from the USB dev
	 */
	{
		usb_device_descr_t * usb_dev_d;
		char devname[20];
		usb_dev_d=usb_get_dev_descr(dip);
		if(usb_dev_d!=NULL){
			usb_get_string_descriptor(dip,USB_LANG_ID,
						  usb_dev_d->iProduct,
						  devname,19);
			devname[19]='\0';
			wacom_debugprint("ProductName is %s\n",devname);
			
		}
		
	}
#endif /* STRINGSAMPLE */

	/* This will create a device node in the appropriate place
	 * under /devices/pci[*]/XXXXXX
	 *
	 * To get an automatic matching link in /dev/XXXX, you need to 
	 * add your own line to /etc/devlink.tab
	 */

	/* XXX I should really check the dev type, and change the
	 * name based on that stuff
	 */
	ddi_create_minor_node(dip, DEVNODE_NAME, S_IFCHR,
			  INST_TO_NODENUM(instance),
			  DEVNODE_TYPE, 0);

	/* There are no PCI-style interrupts, since we dont get to directly
	 * connect to a device.
	 * but an interrupt handler of sorts gets set up in open()
	 */


	ddi_report_dev(dip);

	/* Dont forget to make ddi_remove_minor_node match in detach() */

	return DDI_SUCCESS;

	/* if you add anything to this routine, make ultra-paranoid
	 * checks on return values and return (DDI_FAILURE)
	 * if ANYTHING doesnt go completely 100% okay
	 */
}

/* we dont actualy USE this for anything.
 * At least, not right now.
 */
static	int	wacom_wput(queue_t *q,mblk_t *mblk){
	printf("WPUT called??\n");
	return 0;
}


/* wacom_rput:   precursor to the wacom_rsrv routine.
 * We put stuff on OUR 'STREAM' queue here.
 * Then the rsrv routine moves the stuff over to the UPSTREAM queue,
 * aka the 'stream head', which the user talks to.
 */
static int
wacom_rput(queue_t *q,mblk_t *mp)
{
	wacom_debugprint("^rput called. Placing %p on our queue\n",mp);
	putq(q,mp);
	return 0;
}


/* STREAMS 'service' proceedure, for the read queue.
 * called by the kernel STREAMS scheduler.
 * Basically, just takes all the mouse events we have queued up internally,
 * and shoves them up the stream towards user-level.
 */
static int
wacom_rsrv(queue_t *q)
{
	mblk_t *mq;

	wacom_debugprint("^rsrv called\n");

	while((mq=getq(q)) != NULL){
		if(!canputnext(q)){
			putbq(q,mq);
			break;
		}
		putnext(q,mq);
	}
	return 0;
}


static int wacom_open(queue_t *rq,dev_t *dev, int openflags, int otyp, cred_t *credp){
	int instance=DEVT_TO_INST(*dev);
	usb_pipe_policy_t policy,*policyptr;
	struct wacom_softstate *softstate;

	ASSERT(rq); /* bit of sun paranoia */

	/* Disallow cloning the dev */
	if (openflags != MODOPEN){
		return EINVAL;
	}

	wacom_debugprint("wacom: open for instance %d\n", instance);

	softstate=ddi_get_soft_state(wacom_glob_soft_handle, instance);
	if(softstate==NULL){
		/* This is NOT technically a failure: it is sometimes expected.
		 * It is called "deferred attach"
		 */
		return (ENXIO);
	}

	set_devstate(softstate);

	policyptr=&policy;
	bzero(policyptr,sizeof(policy));

	policyptr->pp_version=USB_PIPE_POLICY_V_0;
	policyptr->pp_timeout_value=USB_PIPE_TIMEOUT;
	policyptr->pp_callback_arg=(void*)softstate;/* same for ALL callbacks*/

	policyptr->pp_callback=wacom_pipe_callback;
	policyptr->pp_exception_callback=wacom_exception_callback;

	/* we do not set these, because there are no ISOC endpoints */
	/*
	policyptr->pp_isoc_callback=wacom_isoc_callback;
	policyptr->pp_isoc_exception_callback=wacom_isoc_exception_callback;
	*/

	/* technically, should look at wMaxPacketSize, and multiply it,
	 * I think. But heck, lets over-allocate
	 */
/*	policyptr->pp_periodic_max_transfer_size=softstate->max_transfer_size;*/
	policyptr->pp_periodic_max_transfer_size=softstate->endpoint.wMaxPacketSize;
/*	policyptr->pp_periodic_max_transfer_size=WACOM_MAX_TRANSFER; */

	/* headerfile seems to hint we shold be able to use
	 * max_transfer_size==0. but that dont work.
	 */
	printf("DEBUG: periodic_max_transfer_size=%d\n",
	        policyptr->pp_periodic_max_transfer_size);

	mutex_enter(&softstate->instance_mutex);
	if(softstate->openflags!=0){
		wacom_debugprint("wacom: device already opened\n");
		mutex_exit(&softstate->instance_mutex);
		return (EBUSY);
	}
	softstate->openflags=WACOM_OPEN_HELD;
	mutex_exit(&softstate->instance_mutex);

	if(usb_pipe_open(softstate->dip,&softstate->endpoint, policyptr,
			 USB_FLAGS_SLEEP,
			 &softstate->usb_pipe_handle) !=USB_SUCCESS){
		cmn_err(CE_WARN,"usb_pipe_open failed");
		softstate->openflags=0;
		return (ENXIO);
	}

	/* Not sure whether this should be after, or before, start_polling */
	softstate->rq=rq;

	/* Waitaminit.. why do I need to set a "policy", if I can
	 * specify a callback thing here??!!
	 */
	if(usb_pipe_start_polling(softstate->usb_pipe_handle,
				  USB_FLAGS_SLEEP,
				  wacom_test_callback,NULL) != USB_SUCCESS)
	{
		cmn_err(CE_WARN,"start_polling FAILED");
		softstate->openflags=0;
		return ENXIO;
	}

	wacom_debugprint("wacom_open called start_polling successfully\n");

	/* Kinda redundant since we can just do ddi_get_soft_state,
	 * but this is the "traditional" streams way of doing things
	 */
	rq->q_ptr = WR(rq)->q_ptr = softstate;
	qprocson(rq);

	

   	return 0;
}


static int wacom_close(queue_t *q, int flag, cred_t *cred_p){
	struct wacom_softstate *softstate;

	/* Okay, extreme paranoia for my first STREAMS dev */
	if(q==NULL) return ENXIO;
	if(q->q_ptr==NULL) return ENXIO;

	qprocsoff(q);

	/* XXX in theory, might want to do additional stuff like
	 *  quntimeout(), qunbufcall(), .... ?
	 */
	softstate=(struct wacom_softstate*)q->q_ptr;

   	wacom_debugprint("wacom: close called\n");
	mutex_enter(&softstate->instance_mutex);
	if(softstate->openflags!= WACOM_OPEN_HELD){
		cmn_err(CE_WARN,"ERROR: wacom_close: pipe not open");
		mutex_exit(&softstate->instance_mutex);
		return ENXIO;
		
	}
	softstate->openflags= WACOM_OPEN_ENDING;
	mutex_exit(&softstate->instance_mutex);

	usb_pipe_stop_polling(softstate->usb_pipe_handle,
			      USB_FLAGS_SLEEP,NULL,NULL);

	if(usb_pipe_close(&softstate->usb_pipe_handle,
			  USB_FLAGS_SLEEP,NULL,NULL ) != USB_SUCCESS)
	{
		wacom_debugprint("ERROR: usb_pipe_close failed?");
		return ENXIO;
	}
	softstate->usb_pipe_handle=NULL;
	softstate->openflags=0;

	q->q_ptr=WR(q)->q_ptr=NULL;

   	return (0);
}

static int wacom_detach(dev_info_t *dip, ddi_detach_cmd_t cmd){
   	int instance;
	struct wacom_softstate *softstate;
   	instance=ddi_get_instance(dip);

	wacom_debugprint("wacom: detach for instance %d\n", instance);

	softstate=ddi_get_soft_state(wacom_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_WARN,"wacom_detach failed: cannot get softstate?");
		return DDI_FAILURE;
	}

	wacom_cleanup(softstate);

	ddi_remove_minor_node(dip, DEVNODE_NAME);

	return (DDI_SUCCESS);
		     
}
